# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap
from lib389._constants import *
from lib389.properties import *
from lib389.tasks import *
from lib389.utils import *
from flask import request
from flask import Blueprint

from .Authentication import authenticateConnection
from .Helpers import jsonResponse, sortEntryAttrs

config_api = Blueprint('config_ui_api', __name__)

"""
This resource handles

- Server Config (cn=config)
- Disk Monitoring
- Access Logging
- Audit Logging
- Errors Logging
- Password Policy
- Security
- SASL Mapping
- LDAPI
- Replication
- Replication Changelog
- Replication Agreements
- Suffix
- Database
- Plugins
"""

#
# o=dmc config attribute entries
#
config_attrs_dn = 'cn=Configuration Attributes,ou=Configuration,o=dmc'
config_dn = 'cn=Config,cn=Configuration Attributes,ou=Configuration,o=dmc'
diskmon_dn = ('cn=Disk Monitoring,cn=config,cn=Configuration Attributes' +
              ',ou=Configuration,o=dmc')
pwp_dn = ('cn=password policy,cn=config,cn=configuration attributes,' +
          'ou=configuration,o=dmc')
errors_log_dn = ('cn=error logging,cn=config,cn=configuration attributes,' +
                 'ou=configuration,o=dmc')
access_log_dn = ('cn=access logging,cn=config,cn=configuration attributes,' +
                 'ou=configuration,o=dmc')
audit_log_dn = ('cn=audit logging,cn=config,cn=configuration attributes,' +
                'ou=configuration,o=dmc')
sasl_dn = ('cn=SASL Mappings,cn=Configuration Attributes,' +
               'ou=Configuration,o=dmc')
ldapi_dn = ('cn=ldapi,cn=config,cn=configuration attributes,' +
                'ou=configuration,o=dmc')
sec_dn = 'cn=Encryption,cn=Configuration Attribute,ou=Configuration,o=dmc'
# Replication
repl_config_dn = ('cn=Configuration Attributes,ou=Configuration,o=dmc')
replica_config_dn = ('cn=replication,cn=configuration attributes,ou=configur' +
                     'ation,o=dmc')
changelog_config_dn = ('cn=changelog,cn=replication,cn=configuration attribu' +
                       'tes,ou=Configuration,o=dmc')
agmt_config_dn = ('cn=agreements,cn=replication,cn=configuration attributes,' +
                  'ou=configuration,o=dmc')
# Suffix/Database
db_dn = "cn=Database,cn=Configuration Attributes,ou=Configuration,o=dmc"
suffix_dn = "cn=Suffix,cn=Configuration Attributes,ou=Configuration,o=dmc"
# Plugins
plugin_dn = "cn=plugins,cn=Configuration Attributes,ou=Configuration,o=dmc"


@config_ui_api.route('/v1/ui/config', methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def getConfigAttrs(ds):
    """
    Search o=dmc for the main configuration attributes the UI is interested in.
    The attributes need to be sorted using the attribute's priority if it's
    set, or sort my alphabetical order (a-z).

    JSON Response
    {
        'config': [attr, attr, attr, ...],
        'diskmon': [attr, attr, attr, ...],
        'access': [attr, attr, attr, ...],
        'audit': [attr, attr, attr, ...],
        'errors': [attr, attr, attr, ...],
        'pwp': [attr, attr, attr, ...],
        'ldapi': [attr, attr, attr, ...],
        'security': [attr, attr, attr, ...],
        'sasl': [attr, attr, attr, ...],
        'replica': [attr, attr, ...],
        'changelog': [attr, attr, ...],
        'agmt': [attr, attr, ...],
        'suffix': [attr, attr, attr, ...],
        'db': [attr, attr, attr, ...]
    }
    """

    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    #
    # Perform the search now so we know how to repsond to the HEAD request
    #
    json_resp = {}

    # Get "all" the configuration form attributes
    try:
        entries = ds.search_s(config_attrs_dn, ldap.SCOPE_SUBTREE,
                              "objectclass=*")
    except ldap.LDAPError as e:
        msg = ('getConfigAttrs - Error: ' + str(e))
        return jsonResponse(msg, 403, request.url)

    for entry in entries:
        config_entry = dict(entry.getJSONEntry())
        if config_entry['dn'].lower() == config_dn:
            json_resp['config'] = sortEntryAttrs(config_entry)
        elif config_entry['dn'].lower() == diskmon_dn:
            json_resp['diskmon'] = sortEntryAttrs(config_entry)
        elif config_entry['dn'].lower() == pwp_dn:
            json_resp['pwp'] = sortEntryAttrs(config_entry)
        elif config_entry['dn'].lower() == access_log_dn:
            json_resp['access'] = sortEntryAttrs(config_entry)
        elif config_entry['dn'].lower() == audit_log_dn:
            json_resp['audit'] = sortEntryAttrs(config_entry)
        elif config_entry['dn'].lower() == errors_log_dn:
            json_resp['errors'] = sortEntryAttrs(config_entry)
        elif config_entry['dn'].lower() == ldapi_dn:
            json_resp['ldapi'] = sortEntryAttrs(config_entry)
        elif config_entry['dn'].lower() == replica_config_dn:
            json_resp['replica'] = sortEntryAttrs(config_entry)
        elif config_entry['dn'].lower() == changelog_config_dn:
            json_resp['changelog'] = sortEntryAttrs(config_entry)
        elif config_entry['dn'].lower() == agmt_config_dn:
            json_resp['agmt'] = sortEntryAttrs(config_entry)
        elif config_entry['dn'].lower() == sec_dn:
            json_resp['security'] = sortEntryAttrs(config_entry)
        elif config_entry['dn'].lower() == sasl_dn:
            json_resp['sasl'] = sortEntryAttrs(config_entry)
        elif config_entry['dn'].lower() == suffix_dn:
            json_resp['suffix'] = sortEntryAttrs(config_entry)
        elif config_entry['dn'].lower() == db_dn:
            json_resp['db'] = sortEntryAttrs(config_entry)
        else:
            # A new config section under cn=configuration attr?
            pass

    #
    # How fulfill the HEAD and GET request
    #
    if request.method == 'HEAD':
        #
        # Return HEAD response - if we get here all is well
        #
        return jsonResponse("", 200, request.url)

    if request.method == 'GET':
        # Everything is already done, just return the JSON response
        return jsonResponse(json_rep, 200, request.url)


@config_ui_api.route('/v1/ui/config/vals', methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def getConfigValues(ds):
    """
    Search the server config for a particular set of attributes from cn=config.

    We do not get the suffix, replication, sasl mapping, and plugin values.
    This is done separately.   We still get the replication changelog values,
    as this is global and not confined to a "suffix/backend".

    JSON Request
    {
        'config': [attr, attr, attr, ...],
        'diskmon': [attr, attr, attr, ...],
        'access': [attr, attr, attr, ...],
        'audit': [attr, attr, attr, ...],
        'errors': [attr, attr, attr, ...],
        'pwp': [attr, attr, attr, ...],
        'ldapi': [attr, attr, attr, ...],
        'security': [attr, attr, attr, ...],
        'changelog': [attr, attr, ...],
        'db': [attr, attr, attr, ...],
    }

    JSON Response
    {
        'config': {'attr': [value, ...],
                   'attr': [value, ...]},
        'diskmon': {'attr': [value, ...],
                    'attr': [value, ...]},
        'access': {'attr': [value, ...],
                   'attr': [value, ...]},
        'audit': {'attr': [value, ...],
                  'attr': [value, ...]},
        'errors': {'attr': [value, ...],
                   'attr': [value, ...]},
        'pwp': {'attr': [value, ...],
                'attr': [value, ...]},
        'ldapi': {'attr': [value, ...],
                  'attr': [value, ...]},
        'security': {'attr': [value, ...],
                     'attr': [value, ...]},
        'changelog': {'attr': [value, ...],
                      'attr': [value, ...]},
        'db': {'attr': [value, ...],
               'attr': [value, ...]}
    }
    """

    #
    # We need to query multiple entries to gather all the config
    #
    # cn=config
    # cn=changelog5,cn=config
    # cn=encryption,cn=config
    # cn=config,cn=ldbm database,cn=plugins,cn=config
    #

    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    #
    # Gather the values now so we can handle the HEAD request
    #
    json_resp = {}
    json_req = request.get_json(force=True)
    if not json_req:
        return jsonResponse('Request is not a valid JSON representation',
                            403, request.url)

    # Start with cn=config
    attr_list = json_req['config']
    try:
        entry = ds.search_s("cn=config", ldap.SCOPE_BASE, "objectclass=*",
                            attr_list)
        json_resp['config'] = sortEntryAttrs(dict(entry[0].getJSONEntry()))
    except ldap.LDAPError as e:
        msg = ('getConfigAttrs - Error: ' + str(e))
        return jsonResponse(msg, 403, request.url)

    # cn=changelog5,cn=config
    attr_list = json_req['changelog']
    try:
        entry = ds.search_s("cn=changelog5,cn=config", ldap.SCOPE_BASE,
                            "objectclass=*", attr_list)
        json_resp['changelog'] = sortEntryAttrs(dict(entry[0].getJSONEntry()))
    except ldap.LDAPError as e:
        # No changelog - ignore
        json_resp['changelog'] = None

    # cn=encryption,cn=config
    attr_list = json_req['security']
    try:
        entry = ds.search_s("cn=encryption,cn=config", ldap.SCOPE_BASE,
                            "objectclass=*", attr_list)
        json_resp['security'] = sortEntryAttrs(dict(entry[0].getJSONEntry()))
    except ldap.LDAPError as e:
        # No security - ignore
        json_resp['security'] = None

    # cn=config,cn=ldbm database,cn=plugins,cn=config
    attr_list = json_req['db']
    try:
        entry = ds.search_s("cn=config,cn=ldbm database,cn=plugins,cn=config",
                            ldap.SCOPE_BASE, "objectclass=*", attr_list)
        json_resp['db'] = sortEntryAttrs(dict(entry[0].getJSONEntry()))
    except ldap.LDAPError as e:
        # No security - ignore
        json_resp['db'] = None

    #
    # Now process the HEAD and GET requests
    #
    if request.method == 'HEAD':
        #
        # Return HEAD response - if we got here, we found some config values
        #
        return jsonResponse("", 200, request.url)

    if request.methos == 'GET':
        #
        # Return our response
        #
        return jsonResponse(json_resp, 200, request.url)


#
# Get Suffix Replication Values
#
@config_ui_api.route('/v1/ui/repl/<suffix>',
                     methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def getReplValues(ds, suffix):
    '''
    Get the replication values for a single suffix.  Gather as much replication
    data as possible so we don't have to keep opening new connections for
    things like the changelog or replication agreements.

    We are expecting a JSON
    request with a list of attributes to retrieve.

    {
        'replica': [attr, attr, ...],
        'agmt': [attr, attr, ...]
    }

    JSON Response
    {
        'replica': {'attr': [value, value], 'attr': [value], ...},
        'agmts': [
                      {'agmt': AGMT_NAME
                       'attrs':  {'attr': [value, value],
                                  'attr': [value],
                                   ...}
                      },
                      {'agmt': AGMT_NAME
                       'attrs':  {'attr': [value, value],
                                  'attr': [value],
                                   ...}
                      },
                 ]
    }
    '''

    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    #
    # Perform all the searches now so we can have a proper HEAD response
    #
    replica_dn = ''
    json_resp = {}
    json_req = request.get_json(force=True)
    if not json_req:
        return jsonResponse('Request is not a valid JSON representation',
                            403, request.url)

    replica_attrs = json_req['replica']
    agmt_attrs = json_req['agmt']

    #
    # Get the replica attr values.  Find the replica config based on the suffix
    #
    filter_str = ("(&(objectclass=nsDS5Replica)(nsDS5Replica=" + suffix + "))")
    try:
        repl_entry = ds.search_s("cn=config", ldap.SCOPE_SUBTREE, filter_str,
                                 replica_attrs)
        if repl_entry:
            replica_dn = repl_entry[0].dn
            json_resp['replica'] = dict(repl_entry[0].data)
        else:
            json_resp['replica'] = None
    except ldap.LDAPError as e:
        msg = ('GetReplValues - Error: ' + str(e))
        return jsonResponse(msg, 403, request.url)

    #
    # Get all the agmt values
    #

    filter_str = "(objectclass=nsDS5ReplicationAgreement"
    agmt_attrs.append('cn')
    try:
        agmt_entries = ds.search_s(replica_dn, ldap.SCOPE_SUBTREE, filter_str,
                                   agmt_attrs)
        if agmt_entries is None:
            json_resp['agmts'] = None
    except ldap.LDAPError as e:
        msg = ('GetReplValues - Error: ' + str(e))
        return jsonResponse(msg, 403, request.url)

    # Process each agmt
    for agmt in agmt_entries:
        json_resp['agmts'].append({'agmt': agmt.getValue('cn'),
                                   'attrs': dict(repl_entry.data)})

    # Sort agmts by agmt name(if not None)
    if json_resp['agmts']:
        json_resp['agmts'].sort(key=operator.itemgetter('agmt'))

    #
    # Now we can handle both HEAD and GET as everything is done
    #
    if request.method == 'HEAD':
        #
        # Return HEAD response
        #
        if agmt_entries is None and cl_entry is None and repl_entry is None:
            return jsonResponse("", 404, request.url)
        else:
            return jsonResponse("", 200, request.url)

    if request.methos == 'GET':
        #
        # Return our response
        #
        return jsonResponse(json_resp, 200, request.url)


# Get sasl mapping values
@config_ui_api.route('/v1/ui/sasl', methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def getSaslValues(ds):
    """
    Get the SASL Mappings from a Directory Server
    {
        'sasl mapping name': {
                                  attr: [value, value, ...],
                                  attr: [value, value, ...]
                             },
        'sasl mapping name2': {
                                   attr: [value, value,...],
                                   attr: [value, value, ...]
                              }
    }
    """

    json_resp = {}
    json_req = request.get_json(force=True)
    if not json_req:
        return jsonResponse('Request is not a valid JSON representation',
                            403, request.url)

    attr_list = json_req
    attr_list.append('cn')
    try:
        entry = ds.search_s("cn=sasl,cn=config", ldap.SCOPE_SUBTREE,
                            "objectclass=nsSaslMapping", attr_list)
    except ldap.LDAPError as e:
        msg = ('GetSaslValues - Error: ' + str(e))
        return jsonResponse(msg, 403, request.url)

    for entry in entries:
        config_entry = dict(entry.getJSONEntry())
        json_resp[config_entry['cn']] = config_entry
    #
    # Now we can handle both HEAD and GET as everything is done
    #
    if request.method == 'HEAD':
        #
        # Return HEAD response
        #
        if entries is None:
            return jsonResponse("", 404, request.url)
        else:
            return jsonResponse("", 200, request.url)

    if request.methos == 'GET':
        #
        # Return our response
        #
        return jsonResponse(json_resp, 200, request.url)


# Get Plugin Values
@config_ui_api.route('/v1/ui/plugins', methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def getPluginVals(ds):
    '''
    Find all the plugins...

    JSON Response
    {
       'plugins': [
                   {'plugin': NAME,
                    'attrs': {'attr': [value, value],
                              'attr': [value, value]},
                   {'plugin': NAME,
                    'attrs': {'attr': [value, value],
                              'attr': [value, value]}
                  ]
    }
    '''
    try:
        entry = ds.search_s("cn=config", ldap.SCOPE_SUBTREE,
                            "objectclass=nsSlapdPlugin")
    except ldap.LDAPError as e:
        msg = ('GetSaslValues - Error: ' + str(e))
        return jsonResponse(msg, 403, request.url)

    for entry in entries:
        config_entry = dict(entry.getJSONEntry())
        json_resp['plugins'].append({'plugin': config_entry['cn'],
                                     'attrs': dict(config_entry)})

    # Sort the plugins by name
    json_resp['plugins'].sort(key=operator.itemgetter('plugin'))

    #
    # Now we can handle both HEAD and GET as everything is done
    #
    if request.method == 'HEAD':
        #
        # Return HEAD response
        #
        if entries is None:
            return jsonResponse("", 404, request.url)
        else:
            return jsonResponse("", 200, request.url)

    if request.methos == 'GET':
        #
        # Return our response
        #
        return jsonResponse(json_resp, 200, request.url)