import ldap
from ldap.controls.readentry import PostReadControl
from lib389._constants import *
from lib389.properties import *
from lib389.tasks import *
from lib389.utils import *
from flask import request, jsonify
from flask import Blueprint

from .Authentication import authenticateConnection
from .Helpers import jsonResponse

index_api = Blueprint('index_api', __name__)

"""
Index Representation::

    {
        'dn': Config DN,
        'index': ATTR,
        'systemIndex': true/false,
        'indexType': [ type, type, ... ]
        'matchingRule': [ mr, mr, ... ]
    }
"""
INDEX = 'cn'
SYSTEM = 'nsSystemIndex'
TYPE = 'nsIndexType'
MR = 'nsMatchingRule'

pr = PostReadControl(criticality=True, attrList=['*'])


def jsonReturnIndexes(indexes, status_code, link):
    # Return our index reps
    results = []
    for index in indexes:
        index_rep = {}
        index_rep['dn'] = index.dn
        index_rep['index'] = index.getValue(INDEX)
        index_rep['systemIndex'] = index.getValue(SYSTEM)
        index_rep['intexType'] = index.getValues(TYPE)
        index_rep['matchingRule'] = index.getValues(MR)
        results.append(dict(index_rep))

    return jsonResponse(results, status_code, link)


@index_api.route('/v1/index', methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def getIndexes(ds):
    """
    List all the DN's of every index
    """

    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    entries = None
    DN = ('cn=ldbm database,cn=plugins,cn=config')
    try:
        entries = ds.search_s(DN, ldap.SCOPE_SUBTREE, '(objectclass=nsIndex)')
    except ldap.LDAPError as e:
        pass

    if request.method == 'HEAD':
        #
        # Return the header
        #
        if entries is None:
            return jsonResponse("", 404, request.url)
        return jsonResponse("", 200, len(entries))

    if request.method == 'GET':
        #
        # Search for every index in the config
        #
        if entries is None:
            msg = ('Failed to find any indexes: ' + e.message['desc'])
            return jsonResponse(msg, 404, request.url)

        return jsonReturnIndexes(entries, 200, request.url)


@index_api.route('/v1/index/<backend>',
                 methods=['GET', 'POST', 'OPTIONS', 'HEAD'])
@authenticateConnection
def getBackendIndexes(ds, backend):
    """
    List all the DN's of every index under a particular backend
    """

    if is_a_dn(backend):
        # backend is actually a suffix - convert it to its backend name
        backend = suffixToBackend(ds, backend)

    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET POST OPTIONS HEAD'
        return jsonify(msg)

    entries = None
    DN = ('cn=index,cn=%s,cn=ldbm database,cn=plugins,cn=config' % backend)
    try:
        entries = ds.search_s(DN, ldap.SCOPE_SUBTREE, '(objectclass=nsIndex)')
    except ldap.LDAPError as e:
        pass

    if request.method == 'HEAD':
        #
        # Return the header
        #
        # NOTE: I don't think this is a great check, there could legitimately
        # be no indexes.  Why is ds.search_s giving us entries as [] if the
        # base doesn't exist?
        #
        #
        if entries is None or []:
            return jsonResponse("", 404, request.url)
        return jsonResponse("", 200, request.url)

    if request.method == 'GET':
        #
        # Search for all the indexes for a backend
        #
        if entries is None or []:
            return jsonResponse("No such backend", 404, request.url)
        #
        # Generate the response and resource link
        #
        return jsonReturnIndexes(entries, 200, request.url)

    if request.method == 'POST':
        if entries is None or []:
            msg = ('No such backend')
            return jsonResponse(msg, 404, request.url)
        #
        # Reindex all the attributes for this backend
        #
        try:
            ret = ds.tasks.reindex(benamebase=backend, args={TASK_WAIT: True})
            if ret == 0:
                #
                # Success, generate the response and resource link
                #
                msg = ds.tasks.entry.getJSONEntry()
                link = getSrchLink(request.url_root,
                                   ds.tasks.dn,
                                   'base',
                                   'objectclass=top')
                return jsonResponse(msg, 200, link)
            else:
                msg = ('Reindex failed: %d' % (ret))
                return jsonResponse(msg, 403, request.url)
        except ValueError as e:
            msg = ('Reindex failed - value error: %s' % e)
            return jsonResponse(msg, 403, request.url)


@index_api.route('/v1/index/<backend>/<attr>', methods=['GET', 'PUT', 'POST',
                 'PATCH', 'DELETE', 'OPTIONS', 'HEAD'])
@authenticateConnection
def attrIndexResource(ds, backend, attr):
    """
    GET    - Get an attribute's index entry
    POST   - Index the attribute
    PUT    - params {index_types} - Add an index to the backend
    DELETE - Delete index definition
    """

    if is_a_dn(backend):
        # backend is actually a suffix - convert it to its backend name
        backend = suffixToBackend(ds, backend)

    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET PATCH PUT POST DELETE OPTIONS HEAD'
        return jsonify(msg)

    # We can use this entry for GET and PUT
    entry = None
    DN = ('cn=%s,cn=index,cn=%s,cn=ldbm database,cn=plugins,cn=config' %
          (attr, backend))
    try:
        entry = ds.search_s(DN, ldap.SCOPE_BASE, '(objectclass=*)')
    except ldap.LDAPError as e:
        pass

    if request.method == 'HEAD':
        #
        # Return the header
        #
        if entry is None:
            return jsonResponse("", 404, request.url)
        return jsonResponse("", 200, request.url)

    if request.method == 'GET':
        #
        # Search for the attribute index
        #
        if entry:
            return jsonReturnIndexes(entry, 200, request.url)
        else:
            msg = ('Failed to find any indexes')
            return jsonResponse(msg, 404, request.url)

    elif request.method == 'POST':
        if entry is None:
            msg = ('Failed to find any indexes')
            return jsonResponse(msg, 404, request.url)
        try:
            #
            # Reindex an attribute
            #
            ret = ds.tasks.reindex(benamebase=backend,
                                   attrname=str(attr),
                                   args={TASK_WAIT: True})
            if ret == 0:
                #
                # Success, generate the response and resource link
                #
                msg = ds.tasks.entry.getJSONEntry()
                link = getSrchLink(request.url_root,
                                   ds.tasks.dn,
                                   'base',
                                   'objectclass=top')
                return jsonResponse(msg, 200, link)
            else:
                msg = ('Reindex failed: %d' % (ret))
                return jsonResponse(msg, 403, request.url)
        except ValueError as e:
            msg = ('Reindex failed - value error: %s' % e)
            return jsonResponse(msg, 403, request.url)

    elif request.method == 'PUT':
        """
        Add an Index - we use PUT because this should be idempotent
        """
        #
        # parse json
        #
        json_req = request.get_json(force=True)

        if not json_req:
            return jsonResponse('Request is not a valid JSON representation',
                                400, request.url)
        #
        # "types" is required to proceed
        #
        if 'indexType' not in json_req or 'index' not in json_req:
            return jsonResponse('JSON representation is missing "indexType"' +
                                'or "index"', 400, request.url)

        else:
            DN = ('cn=%s,cn=index,cn=%s,cn=ldbm database,cn=plugins,cn=config'
                  % (json_req['index'], backend))

            #
            # Need to get the types/matching rules lists and convert the
            # unicode dict values to strings
            #
            if 'matchingRule' in json_req:
                mr = list(map(str, json_req['matchingRule']))
            else:
                mr = None

            try:
                #
                # Add the index
                #
                index_types = list(map(str, json_req['indexType']))
                msg_id = ds.index.addIndex(suffix=None,
                                           be_name=backend,
                                           attr=str(json_req['index']),
                                           indexTypes=index_types,
                                           matchingRules=mr,
                                           postReadCtrl=True)

                #
                # Generate the response and resource link
                #
                try:
                    _, _, _, resp_ctrls = ds.result3(msg_id)
                except ldap.ALREADY_EXISTS:
                    # This is okay, just need to get the entry for the response
                    link = '%sv1/index/%s/%s' % (request.url_root, backend,
                                                 attr)
                    return jsonResponse(entry[0].getJSONEntry(), 201, link)
                except ldap.LDAPError as e:
                    msg = ('Failed to add index, error: ' + str(e))
                    return jsonResponse(str(e), 403, request.url)

                entry = resp_ctrls[0].entry
                if MR in entry:
                    mr = entry[MR]
                else:
                    mr = ''
                index_rep = {'dn': resp_ctrls[0].dn,
                             'index': entry[INDEX],
                             'systemIndex': entry[SYSTEM],
                             'indexType': entry[TYPE],
                             'matchingRule': mr}
                link = '%sv1/index/%s/%s' % (request.url_root, backend, attr)
                return jsonResponse(index_rep, 201, link)

            except ldap.LDAPError as e:
                msg = ('Failed to add index, error: ' + str(e))
                return jsonResponse(msg, 403, request.url)

    elif request.method == 'PATCH':
        #
        # Modify/patch the index representation
        #
        mod_list = []
        json_rep = request.get_json(force=True)
        for mod in json_rep:
            # Get the op (add, replace, delete) and parse the path
            op = jsonGetPatchOp(mod)
            if op == -1:
                return jsonResponse(('invalid PATCH op (%s)' % mod['op']), 400,
                                    request.url)

            if mod['path'] == 'index':
                mod_attr = INDEX
            elif mod['path'] == 'indexType':
                mod_attr = TYPE
            elif mod['path'] == 'matchingRule':
                mod_attr = MR
            elif mod['path'] == 'systemIndex':
                mod_attr = SYSTEM
            else:
                # Some other attribute?
                mod_attr = mod['path']

            if isinstance(mod['value'], list):
                value = list(map(str, mod['value']))
            else:
                value = str(mod['value'])
            mod_list.append((op, mod_attr, value))

        #
        # Perform the modifications
        #
        try:
            DN = ('cn=%s,cn=index,cn=%s,cn=ldbm database,cn=plugins,cn=config'
                  % (attr, backend))
            msg_id = ds.modify_ext(DN, mod_list, serverctrls=[pr])
        except ldap.LDAPError as e:
            return jsonResponse('Failed to modify resource: ' + str(e), 403,
                                request.url)

        #
        # Generate the response and resource link
        #
        try:
            _, _, _, resp_ctrls = ds.result3(msg_id)
            entry = resp_ctrls[0].entry
            if MR in entry:
                mr = entry[MR]
            else:
                mr = ''

            index_rep = {'dn': resp_ctrls[0].dn,
                         'index': entry[INDEX],
                         'systemIndex': entry[SYSTEM],
                         'indexType': entry[TYPE],
                         'matchingRule': mr}
            link = '%sv1/index/%s/%s' % (request.url_root, backend, attr)
            return jsonResponse(index_rep, 200, link)

        except ldap.LDAPError as e:
            return jsonResponse('Failed to modify resource: ' + str(e), 403,
                                request.url)

    elif request.method == 'DELETE':
        """
        Delete Index
        """
        try:
            DN = ('cn=%s,cn=index,cn=%s,cn=ldbm database,cn=plugins,cn=config'
                  % (attr, backend))
            try:
                ds.delete_s(DN)
            except ldap.NO_SUCH_OBJECT:
                # already deleted?  idempotent - return success
                pass
            except ldap.LDAPError as e:
                msg = ('Failed to delete index, error: ' + e.message['desc'])
                return jsonResponse(msg, 403, request.url)
            msg = 'Delete Complete'
            return jsonResponse(msg, 200, request.url)
        except ldap.LDAPError as e:
            msg = ('Failed to delete index: ' + e.message['desc'])
            return jsonResponse(msg, 403, request.url)
