import ldap
from ldap.controls.readentry import PostReadControl
from lib389._constants import *
from lib389.properties import *
from lib389.tasks import *
from lib389.utils import *
from flask import request, jsonify
from flask import Blueprint

from .Authentication import authenticateConnection
from .Helpers import (jsonResponse, getSrchLink, postEntryToRep, jsonGetPatchOp)

ldap_api = Blueprint('ldap_api', __name__)

"""
LDAP Entry Representation::

    [
        {
           "dn": "DN",
           "attrs": {
                         "ATTR": ["VALUE", "VALUE"],
                         "ATTR": ["VALUE"],
                         ...,
                         ...
                    }
        },
        {
           "dn": "DN",
           "attrs": {
                         "ATTR": ["VALUE", "VALUE"],
                         "ATTR": ["VALUE"]
                         ...,
                         ...
                    }
        }
    ]
"""

SCOPE = {}
SCOPE['sub'] = ldap.SCOPE_SUBTREE
SCOPE['one'] = ldap.SCOPE_ONELEVEL
SCOPE['base'] = ldap.SCOPE_BASE

# Control for returning a representation of an entry after it was
# modified/created.
pr = PostReadControl(criticality=True, attrList=['*'])


def jsonReturnEntries(entries, link):
    """
    Return a JSON representation of ldap entries
    """
    if not entries:
        return jsonResponse('No entries', 404, request.url)

    results = []  # List of entries
    count = 0
    for entry in entries:
        results.append(dict(entry.getJSONEntry()))
        count += 1

    return jsonResponse(results, 200, link)


@ldap_api.route('/v1/dit/<dn>', methods=['GET', 'PUT', 'PATCH', 'POST',
                                         'DELETE', 'OPTIONS', 'HEAD'])
@authenticateConnection
def ldap_req(ds, dn):
    """
    Perform an ldap modify operation::

        GET    = Search  JSON Rep (None)
        PUT    = Add     JSON Rep {'dn': DN, 'attrs': [{'attr': VALUE},
                                                       {'attr': VALUE}, ...]}
        PATCH  = Modify  JSON Rep [{'op': MOD_TYPE, 'path': ATTR,
                                    'value': [VALUE, VALUE,...}]
        POST   = Modrdn  JSON Rep {'dn': DN, 'newrdn': DN, 'newsuperior': DN,
                                   'deleteoldrdn': INT}
        DELETE = Delete  JSON Rep (None)

    """

    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET POST PUT PATCH DELETE OPTIONS HEAD'
        return jsonify(msg)

    # Perform a search on the entry (used by HEAD and PUT)
    try:
        entry = ds.search_s(dn, ldap.SCOPE_BASE, '(objectclass=*)')
    except ldap.LDAPError as e:
        print(str(e))

    if request.method == 'HEAD':
        #
        # Perform HEAD Request
        #
        result = 200
        if not entry:
            result = 404
        return jsonResponse("", result, request.url)

    elif request.method == 'GET':
        #
        # Search just the DN
        #
        try:
            entries = ds.search_s(dn, SCOPE['sub'])
            return jsonReturnEntries(entries, request.url)
        except ldap.LDAPError as e:
            msg = ("Error: " + str(e))
            return jsonResponse(msg, 404, request.url)

    elif request.method == 'PUT':
        """
        Add an entry
        """
        #
        # Parse JSON request
        #
        json_req = request.get_json(force=True)
        if not json_req:
            return jsonResponse('Missing JSON', 401, request.url)

        add_record = []
        for attrlist in json_req['attrs']:
            for attr in attrlist:
                if isinstance(attrlist[attr], list):
                    for val in attrlist[attr]:
                        add_record.append((str(attr), str(val)))
                else:
                    add_record.append((str(attr), str(attrlist[attr])))

        try:
            #
            # Add the entry
            #
            msg_id = ds.add_ext(dn, add_record, serverctrls=[pr])
            # Get the results(and post op entry)
            _, _, _, resp_ctrls = ds.result3(msg_id)
        except ldap.ALREADY_EXISTS:
            # Need to get the entry, in order to return it in the response
            link = getSrchLink(request.url_root, dn, 'base', 'objectclass=*')
            return jsonResponse(entry[0].getJSONEntry(), 201, link)
        except ldap.LDAPError as e:
            return jsonResponse('Failed to add entry: %s' % str(e), 403,
                                request.url)
        msg = postEntryToRep(resp_ctrls[0])
        link = getSrchLink(request.url_root, dn, 'base', 'objectclass=*')
        return jsonResponse(msg, 201, link)

    elif request.method == 'PATCH':
        """
        Modify an entry
        """
        #
        # Parse JSON request
        #
        json_req = request.get_json(force=True)
        mod_list = []
        for mod in json_req:
            # Get the op (add, replace, remove) and parse the path
            op = jsonGetPatchOp(mod)
            if op == -1:
                return jsonResponse(('invalid PATCH op (%s)' % mod['op']), 403,
                                    request.url)

            mod_attr = str(mod['path'])
            value = str(mod['value'])
            mod_list.append((op, mod_attr, value))

        #
        # Perform the modifications
        #
        try:
            msg_id = ds.modify_ext(dn, mod_list, serverctrls=[pr])
        except ldap.LDAPError as e:
            return jsonResponse('Failed to modify resource: ' + str(e), 403,
                                request.url)

        #
        # Generate the response and resource link
        #
        try:
            _, _, _, resp_ctrls = ds.result3(msg_id)
        except ldap.LDAPError as e:
            return jsonResponse('Failed to modify resource: ' + str(e), 403,
                                request.url)

        link = getSrchLink(request.url_root, dn, 'base', 'objectclass=*')
        return jsonResponse(postEntryToRep(resp_ctrls[0]), 200, link)

    elif request.method == 'POST':
        """
        Modrdn::

            {
                'newrdn': NEWRDN
                'newsuperior': NEWSUPERIOR
                'deleteoldrdn': 1 or 0
            }
        """
        #
        # Parse JSON request
        #
        json_req = request.get_json(force=True)
        if not json_req:
            return "NO JSON"

        if 'newrdn' in json_req:
            newrdn = json_req['newrdn']
        else:
            # Error, can not recover
            msg = ('Missing "newrdn" in request')
            return jsonResponse(msg, 403, request.url)

        if 'newsuperior' in json_req:
            newsup = str(json_req['newsuperior'])
        else:
            newsup = None

        if 'deleteoldrdn' in json_req:
            deloldrdn = json_req['deleteoldrdn']
        else:
            deloldrdn = 0

        #
        # Perform modrdn
        #
        try:
            msg_id = ds.rename(dn,
                               newrdn,
                               newsuperior=newsup,
                               delold=deloldrdn,
                               serverctrls=[pr])

            #
            # Generate the response and resource link
            #
            _, _, _, resp_ctrls = ds.result3(msg_id)
            link = getSrchLink(request.url_root, resp_ctrls[0].dn, 'base',
                               'objectclass=top')
            msg = postEntryToRep(resp_ctrls[0])
            return jsonResponse(msg, 200, link)
        except ldap.LDAPError as e:
            msg = ('Failed to modrdn entry (%s): error %s' % (dn, str(e)))
            return jsonResponse(msg, 403, request.url)

    elif request.method == 'DELETE':
        #
        # Delete the entry
        #
        try:
            ds.delete_s(dn)
            return jsonResponse("Entry deleted", 200, request.url)
        except ldap.NO_SUCH_OBJECT:
            pass
        except ldap.LDAPError as e:
            msg = ('Failed to delete entry (%s): error %s' % (dn, str(e)))
            return jsonResponse(msg, 403, request.url)


@ldap_api.route('/v1/dit/<dn>/<scope>', methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def search_S(ds, dn, scope):
    """
    Search it: Scope
    """
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    if request.method == 'HEAD':
        #
        # Perform HEAD Request
        #
        result = 200
        entries = []
        try:
            entries = ds.search_s(dn, SCOPE[scope])
        except ldap.LDAPError as e:
            result = 404
        return jsonResponse("", result, request.url)
    else:
        try:
            entries = ds.search_s(dn, SCOPE[scope])
            return jsonReturnEntries(entries, request.url)
        except ldap.LDAPError as e:
            msg = ("Error: " + str(e))
            return jsonResponse(msg, 404, request.url)


@ldap_api.route('/v1/dit/<dn>/<scope>/<filter_str>', methods=['GET',
                                                              'OPTIONS',
                                                              'HEAD'])
@authenticateConnection
def search_SF(ds, dn, scope, filter_str):
    """
    Search it: Scope, Filter
    """
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    if request.method == 'HEAD':
        #
        # Perform HEAD Request
        #
        result = 200
        entries = []
        try:
            entries = ds.search_s(dn, SCOPE[scope], filter_str)
        except ldap.LDAPError as e:
            result = 404
        return jsonResponse("", result, request.url)

    elif request.method == 'GET':
        try:
            entries = ds.search_s(dn, SCOPE[scope], filter_str)
            return jsonReturnEntries(entries, request.url)
        except ldap.LDAPError as e:
            msg = ("Error: " + str(e))
            return jsonResponse(msg, 404, request.url)


@ldap_api.route('/v1/dit/<dn>/<scope>/<filter_str>/<attrs>',
                methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def search_SFA(ds, dn, scope, filter_str, attrs):
    """
    Search it:  Scope, Filter, Attrs (comma separated)
    """
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    if request.method == 'HEAD':
        #
        # Perform HEAD Request
        #
        result = 200
        entries = []
        try:
            entries = ds.search_s(dn, SCOPE[scope], filter_str)
        except ldap.LDAPError as e:
            result = 404
        return jsonResponse("", result, request.url)

    elif request.method == 'GET':
        try:
            alist = list(map(str, attrs.split(',')))
            entries = ds.search_s(dn,
                                  SCOPE[scope],
                                  filter_str,
                                  attrlist=alist)
            return jsonReturnEntries(entries, request.url)
        except ldap.LDAPError as e:
            msg = ("Error: " + str(e))
            return jsonResponse(msg, 404, request.url)


@ldap_api.route('/v1/dit/rootdse',
                methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def search_rootdse(ds):
    """
    Get the root DSE entry
    """
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    try:
        entries = ds.search_s("", ldap.SCOPE_BASE, "objectclass=top")
        result = 200
    except ldap.LDAPError as e:
        result = 404

    if request.method == 'HEAD':
        #
        # Perform HEAD Request
        #
        return jsonResponse("", result, request.url)

    elif request.method == 'GET':
        if result == 200:
            return jsonReturnEntries(entries, request.url)
        else:
            msg = ("Error: " + str(e))
            return jsonResponse(msg, 404, request.url)


@ldap_api.route('/v1/dit/rootdse/<attrs>',
                methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def search_rootdse_attrs(ds, attrs):
    """
    Get the root DSE entry with requested attributes
    """
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    if request.method == 'HEAD':
        #
        # Perform HEAD Request
        #
        result = 200
        entries = []
        try:
            entries = ds.search_s("", SCOPE[base], "objectclass=top")
        except ldap.LDAPError as e:
            result = 404
        return jsonResponse("", result, request.url)

    elif request.method == 'GET':
        try:
            alist = list(map(str, attrs.split(',')))
            entries = ds.search_s("",
                                  ldap.SCOPE_BASE,
                                  "objectclass=top",
                                  attrlist=alist)
            return jsonReturnEntries(entries, request.url)
        except ldap.LDAPError as e:
            msg = ("Error: " + str(e))
            return jsonResponse(msg, 404, request.url)
