# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap
import binascii
from lib389 import Entry
from ldap.controls.readentry import PostReadControl
from lib389._constants import *
from lib389.properties import *
from lib389.tasks import *
from lib389.utils import *
from flask import request, jsonify
from flask import Blueprint

from .Authentication import authenticateConnection
from .Helpers import jsonResponse

repl_api = Blueprint('repl_api', __name__)

pr = PostReadControl(criticality=True, attrList=['*'])

TYPE = ['Unknown', 'Primary', 'Read-Only (Hub/Consumer)', 'master']


def decodeEntryBase64(data):
    b64 = binascii.b2a_base64(data)
    if b64[-1] == '\n':
        # Strip off the newline
        b64 = b64[:-1]

    d64 = b64.decode('base64')
    try:
        d64.decode('ascii')
    except UnicodeDecodeError:
        # Just return the base64 string since its not ascii
        d64 = b64

    return d64


def jsonReturnReplicas(replicas, status_code, link):
    """
    Take replica entries and convert them to the replica representation
    """
    results = []
    for replica in replicas:
        repl_rep = {}
        repl_rep['dn'] = replica.dn
        repl_rep['ReplicaRoot'] = replica.getValue(REPL_ROOT)
        repl_rep['ReplicaRole'] = \
            TYPE[int(replica.getValue(REPL_TYPE))]
        repl_rep['ReplicaType'] = replica.getValue(REPL_TYPE)
        if not replica.getValue(REPL_FLAGS):
            repl_rep['ReplicaFlags'] = TYPE[0]
        else:
            repl_rep['ReplicaFlags'] = \
                int(replica.getValue(REPL_FLAGS))
        repl_rep['ReplicaID'] = replica.getValue(REPL_ID)
        repl_rep['ReplicaPurgeDelay'] = replica.getValue(REPL_PURGE_DELAY)
        repl_rep['ReplicaBindDN'] = replica.getValue(REPL_BINDDN)
        repl_rep['ReplicaBindDNGroup'] = replica.getValue(REPL_BIND_GROUP)
        repl_rep['ReplicaBindDNGroupCheckInterval'] = \
            replica.getValue(REPL_BIND_GROUP_INTERVAL)
        repl_state = replica.getValue(REPL_STATE)  # base 64
        if repl_state:
            repl_state = decodeEntryBase64(repl_state)
        repl_rep['ReplicaState'] = str(repl_state)
        repl_rep['ReplicaName'] = replica.getValue(REPL_NAME)
        backoff_min = replica.getValue(REPL_RETRY_MIN)
        if not backoff_min:
            backoff_min = '3'
        repl_rep['ReplicaBackoffMin'] = backoff_min
        backoff_max = replica.getValue(REPL_RETRY_MAX)
        if not backoff_max:
            backoff_max = '300'
        repl_rep['ReplicaBackoffMax'] = backoff_max
        ptimeout = replica.getValue(REPL_PROTOCOL_TIMEOUT)
        if not ptimeout:
            ptimeout = '120'
        repl_rep['ReplicaProtocolTimeout'] = ptimeout
        repl_rep['ReplicaReferral'] = replica.getValues(REPL_REF)
        repl_rep['ReplicaTombstonePurgeInterval'] = \
            replica.getValue(REPL_TOMBSTONE_PURGE_INTERVAL)
        repl_rep['ReplicaLegacyConsumer'] = \
            replica.getValue(REPL_LEGACY_CONS)
        repl_rep['ReplicaCleanRUV'] = replica.getValue(REPL_CLEAN_RUV)
        repl_rep['ReplicaAbortCleanRUV'] = replica.getValue(REPL_CLEAN_RUV)
        repl_rep['ReplicaChangeCount'] = replica.getValue(REPL_COUNT_COUNT)
        repl_rep['ReplicaReapActive'] = replica.getValue(REPL_REAP_ACTIVE)
        repl_rep['ReplicaPreciseTombstonePurging'] = \
            replica.getValue(REPL_PRECISE_PURGE)

        results.append(dict(repl_rep))

    return jsonResponse(results, status_code, link)


def jsonReturnReplAgmts(agmts, status_code, link):
    '''
    Return JSON presentation of all replication agreements
    '''
    results = []
    for agmt in agmts:
        agmt_rep = {}
        agmt_rep['dn'] = agmt.dn
        agmt_rep['ReplicaName'] = agmt.getValue(RA_NAME)
        agmt_rep['ReplicaRoot'] = agmt.getValue(REPL_ROOT)
        agmt_rep['ReplicaBindMethod'] = agmt.getValue(AGMT_BIND_METHOD)
        agmt_rep['ReplicaMaxCSN'] = agmt.getValue(AGMT_MAXCSN)
        agmt_rep['ReplicaRUV'] = agmt.getValues(REPL_RUV)
        agmt_rep['ReplicaBindDN'] = agmt.getValue(REPL_BINDDN)
        agmt_rep['ReplicaBindCredentials'] = agmt.getValue(AGMT_CRED)
        agmt_rep['ReplicaHost'] = agmt.getValue(AGMT_HOST)
        agmt_rep['ReplicaPort'] = agmt.getValue(AGMT_PORT)
        agmt_rep['ReplicaTransportInfo'] = agmt.getValue(AGMT_TRANSPORT_INFO)
        agmt_rep['ReplicaEnabled'] = agmt.getValue(AGMT_ENABLED)
        agmt_rep['ReplicaConnTimeout'] = agmt.getValue(AGMT_TIMEOUT)
        agmt_rep['ReplicaRefresh'] = agmt.getValue(AGMT_REFRESH)
        agmt_rep['ReplicaBusyWaitTime'] = agmt.getValue(AGMT_BUSY_WAIT_TIME)
        agmt_rep['ReplicaSessionPauseTime'] = \
            agmt.getValue(AGMT_SESSION_PAUSE_TIME)
        agmt_rep['ReplicaFlowControlWindow'] = agmt.getValue(AGMT_FLOW_WINDOW)
        agmt_rep['ReplicaFlowControlPause'] = agmt.getValue(AGMT_FLOW_PAUSE)
        agmt_rep['ReplicatedAttributeList'] = agmt.getValue(AGMT_ATTR_LIST)
        agmt_rep['ReplicatedAttributeListTotal'] = \
            agmt.getValue(AGMT_ATTR_LIST_TOTAL)
        agmt_rep['ReplicaSchedule'] = agmt.getValue(AGMT_SCHEDULE)
        agmt_rep['ReplicaStripAttrs'] = agmt.getValue(AGMT_STRIP_ATTRS)
        agmt_rep['ReplicaReapActive'] = agmt.getValue(REPL_REAP_ACTIVE)
        backoff_min = agmt.getValue(REPL_RETRY_MIN)
        if not backoff_min:
            backoff_min = '3'
        agmt_rep['ReplicaBackoffMin'] = backoff_min
        backoff_max = agmt.getValue(REPL_RETRY_MAX)
        if not backoff_max:
            backoff_max = '300'
        agmt_rep['ReplicaBackoffMax'] = backoff_max

        # Status attributes
        agmt_rep['ReplicaLastUpdateStart'] = agmt.getValue(AGMT_UPDATE_START)
        agmt_rep['ReplicaLastUpdateEnd'] = agmt.getValue(AGMT_UPDATE_END)
        agmt_rep['ReplicaLastUpdateStatus'] = agmt.getValue(AGMT_UPDATE_STATUS)
        agmt_rep['ReplicaUpdateInProgress'] = \
            agmt.getValue(AGMT_UPDATE_IN_PROGRESS)
        agmt_rep['ReplicaLastInitStart'] = agmt.getValue(AGMT_INIT_START)
        agmt_rep['ReplicaLastInitEnd'] = agmt.getValue(AGMT_INIT_END)
        agmt_rep['ReplicaLastInitStatus'] = agmt.getValue(AGMT_INIT_STATUS)
        changes_sent = agmt.getValue(AGMT_CHANGES_SINCE_STARTUP)  # base64
        if changes_sent:
            changes_sent = decodeEntryBase64(changes_sent)
        agmt_rep['ReplicaChangesSentSinceStartup'] = changes_sent

        if agmt.hasValue('objectcass', RA_WINDOWS_OBJECTCLASS_VALUE):
            # Winsync specific attributes
            agmt_rep['WinsyncDirectoryReplicaSubtree'] = \
                agmt.getValue(AGMT_WINSYNC_REPLICA_SUBTREE)
            agmt_rep['WindowsReplicaSubtree'] = \
                agmt.getValue(AGMT_WINSYNC_SUBTREE)
            agmt_rep['WinsyncNewWinUserSyncEnabled'] = \
                agmt.getValue(AGMT_WINSYNC_NEW_USERSYNC)
            agmt_rep['WinsyncNewWinGroupSyncEnabled'] = \
                agmt.getValue(AGMT_WINSYNC_NEW_GROUP_SYNC)
            agmt_rep['WinsyncWindowsDomain'] = \
                agmt.getValue(AGMT_WINSYNC_DOMAIN)
            agmt_rep['WinsyncDirsyncCookie'] = \
                agmt.getValue(AGMT_WINSYNC_COOKIE)
            agmt_rep['WinsyncInterval'] = \
                agmt.getValue(AGMT_WINSYNC_INTERVAL)
            agmt_rep['WinsyncOneWaySync'] = \
                agmt.getValue(AGMT_WINSYNC_ONE_WAY_SYNC)
            agmt_rep['WinsyncMoveAction'] = \
                agmt.getValue(AGMT_WINSYNC_MOVE_ACTION)
            agmt_rep['WinsyncWindowsFilter'] = \
                agmt.getValue(AGMT_WINSYNC_WIN_FILTER)
            agmt_rep['WinsyncDirectoryFilter'] = \
                agmt.getValue(AGMT_WINSYNC_DIR_FILTER)
            agmt_rep['WinsyncSubtreePair'] = \
                agmt.getValue(AGMT_WINSYNC_SUBTREE_PAIR)

        results.append(dict(agmt_rep))

    return jsonResponse(results, status_code, link)


def jsonReturnChangelog(changelog, status_code, link):
    """
    Return changelog representation
    """
    cl_rep = {}
    cl_rep['dn'] = changelog.dn
    cl_rep['ChangelogDir'] = changelog.getValue(CL_DIR)
    cl_rep['ChangelogMaxEntries'] = changelog.getValue(CL_MAX_ENTRIES)
    cl_rep['ChangelogMaxAge'] = changelog.getValue(CL_MAXAGE)
    cl_rep['ChangelogCompactDBInterval'] = \
        changelog.getValue(CL_COMPACT_INTERVAL)
    cl_rep['ChangelogTrimInterval'] = changelog.getValue(CL_TRIM_INTERVAL)
    cl_rep['ChangelogMaxConcurrentWrites'] = \
        changelog.getValue(CL_CONCURRENT_WRITES)
    cl_rep['ChangelogEncryptionAlgorithm'] = changelog.getValue(CL_ENCRYPT_ALG)
    cl_rep['ChangelogSymmetricKey'] = changelog.getValue(CL_SYM_KEY)

    return jsonResponse(cl_rep, status_code, link)


def jsonReturnAgmtStatus(agmts):
    if len(agmts) == 1:
        agmt = agmts[0]
        json_rep = {'agreement': agmt[0],
                    'status': agmt[1]}
        jsonResponse(json_rep, 200, request.url)
    else:
        json_rep = []  # List of agreements
        for result in agmts:
            status = {'agreement': result[0],
                      'status': result[1]}
            json_rep.append(dict(status))
        jsonResponse(json_rep, 200, request.url)


def doReplAgmtPOST(ds, agmts):
        #
        # Parse and validate json request
        #
        json_req = request.get_json(force=True)
        if not json_req:
            return jsonResponse('Request is not a valid JSON representation',
                                403, request.url)
        if 'action' in json_req:
            action = str(json_req['action']).lower()
            if action != 'initialize' and action != 'send_updates' and \
               action != 'enable' and action != 'disable':
                return jsonResponse('Request action is unknown: %s' % action,
                                    403, request.url)
        else:
            return jsonResponse('Request is missing agreement action',
                                403, request.url)
        if 'target' in json_req:
            target = str(json_req['target']).lower()
            if target != 'all' and target != 'winsync' and target != 'ds':
                return jsonResponse('Request has valid target: %s' % target,
                                    403, request.url)
        else:
            target = 'all'

        # Determine if this is one or many
        if len(agmts) == 1:
            single_agmt = True
        else:
            single_agmt = False

        failure_list = []

        #
        # Enable/Disable repl agreements
        #
        if action == 'enable' or action == 'disable':
            if action == 'enable':
                state = 'on'
                msg = 'enabled'
            else:
                state = 'off'
                msg = 'disabled'
            if target == 'ds':
                filter_str = '(objectclass=%s)' % RA_OBJECTCLASS_VALUE
            elif target == 'winsync':
                filter_str = '(objectclass=%s)' % RA_WINDOWS_OBJECTCLASS_VALUE

            for agmt in agmts:
                dn = agmt.dn
                if target == 'ds':
                    if not agmt.hasValue('objectcass', RA_OBJECTCLASS_VALUE):
                        continue
                elif target == 'winsync':
                    if not agmt.hasValue('objectcass',
                                         RA_WINDOWS_OBJECTCLASS_VALUE):
                        continue

                # If we get here we want to modify this agmt
                msg_id = ds.modify_ext(dn, [(ldap.MOD_REPLACE,
                                             AGMT_ENABLED,
                                             state)], serverctrls=[pr])
                try:
                    _, _, _, resp_ctrls = ds.result3(msg_id)
                    agmt_entry = Entry(resp_ctrls[0].dn)
                    agmt_entry.data = resp_ctrls[0].entry
                except:
                    # Create a list of agmts that were not updated
                    failure_list.append(dn)

            if single_agmt:
                if len(failure_list) > 0:
                    return jsonReponse('Failed to update agreement', 403,
                                       request.url)
                else:
                    name = agmt_entry.getValue('cn')
                    suffix = agmt.getValue(REPL_ROOT)
                    link = ('%s/v1/replication/%s/%s' % (request.url_root,
                                                         suffix, name))
                    return jsonReturnReplAgmts([agmt_entry], 200, link)
            else:
                if len(failure_list) > 0:
                    resp_msg = {'result_status': 'Failed to update agreements',
                                'agreements': failure_list}
                    return jsonResponse(resp_msg, 200, link)
                else:
                    link = getSrchLink(request.url_root, 'cn=config', 'sub',
                                       filter_str)
                    resp_msg = 'All replication agreements have been %s' % msg
                    return jsonResponse(resp_msg, 200, link)

        #
        # Initialize repl agreements
        #
        if action == 'initialize':
            for agmt in agmts:
                dn = agmt.dn
                if target == 'ds':
                    if not agmt.hasValue('objectcass', RA_OBJECTCLASS_VALUE):
                        continue
                elif target == 'winsync':
                    if not agmt.hasValue('objectcass',
                                         RA_WINDOWS_OBJECTCLASS_VALUE):
                        continue

                # If we get here we want to initialize this agmt
                msg_id = ds.modify_ext(dn, [(ldap.MOD_REPLACE,
                                             AGMT_REFRESH,
                                             'start')], serverctrls=[pr])
                try:
                    _, _, _, resp_ctrls = ds.result3(msg_id)
                    agmt_entry = Entry(resp_ctrls[0].dn)
                    agmt_entry.data = resp_ctrls[0].entry
                except ldap.LDAPError as e:
                    # Create a list of agmts that were not updated
                    failure_list.append(dn)

            if single_agmt:
                if len(failure_list) > 0:
                    return jsonResponse(
                        'Failed to initialize replicaton agreement (%s)' %
                        (dn), 403, request.url)
                else:
                    name = agmt_entry.getValue('cn')
                    suffix = agmt.getValue(REPL_ROOT)
                    link = ('%s/v1/replication/%s/%s' % (request.url_root,
                                                         suffix, name))
                    return jsonReturnReplAgmts([agmt_entry], 200, link)
            else:
                if len(failure_list) > 0:
                    resp_msg = {'result_status':
                                'Failed to initialize all agreements',
                                'agreements': failure_list}
                    return jsonResponse(resp_msg, 403, link)
                else:
                    link = getSrchLink(request.url_root, 'cn=config', 'sub',
                                       filter_str)
                    resp_msg = 'All agreements have been initialized'
                    return jsonResponse(resp_msg, 200, link)

        #
        # Send updates for all agreements
        #
        if action == 'send_updates':
            for agmt in agmts:
                dn = agmt.dn
                if target == 'ds':
                    if not agmt.hasValue('objectcass', RA_OBJECTCLASS_VALUE):
                        continue
                elif target == 'winsync':
                    if not agmt.hasValue('objectcass',
                                         RA_WINDOWS_OBJECTCLASS_VALUE):
                        continue

                # If we get here we want to "send_updates" to this agmt.  Add
                # and delete the agtm schedule attribute
                try:
                    ds.modify_s(dn, [(ldap.MOD_ADD, AGMT_SCHEDULE, '*')])
                except ldap.LDAPError as e:
                    failure_list.append(dn)
                    continue

                # Now remove that attribute/value
                msg_id = ds.modify_ext(dn,
                                       [(ldap.MOD_DELETE,
                                         AGMT_SCHEDULE,
                                         '*')],
                                       serverctrls=[pr])
                try:
                    _, _, _, resp_ctrls = ds.result3(msg_id)
                    agmt_entry = Entry(resp_ctrls[0].dn)
                    agmt_entry.data = resp_ctrls[0].entry
                except:
                    # Create a list of agmt that were not updated
                    failure_list.append(dn)

            if single_agmt:
                if len(failure_list) > 0:
                    return jsonResponse(
                        'Failed to send updates for agreement (%s): %s' %
                        (dn, str(e)), 403, request.url)
                else:
                    name = agmt_entry.getValue('cn')
                    suffix = agmt.getValue(REPL_ROOT)
                    link = ('%s/v1/replication/%s/%s' % (request.url_root,
                                                         suffix, name))
                    return jsonReturnReplAgmts([agmt_entry], 200, link)
            else:
                if len(failure_list) > 0:
                    resp_msg = {'result_status': 'Failed to send updates',
                                'agreements': failure_list}
                    return jsonResponse(resp_msg, 403, request.url)
                else:
                    link = getSrchLink(request.url_root, 'cn=config', 'sub',
                                       filter_str)
                    resp_msg = 'All replication agreements have sent updates'
                    return jsonResponse(resp_msg, 200, link)


@repl_api.route('/v1/replication/changelog',
                methods=['GET', 'PATCH', 'PUT', 'DELETE', 'OPTIONS', 'HEAD'])
@authenticateConnection
def changeLog(ds):
    """
    Changelog Resource - Create, modify, and delete changelog

    Changelog Representation::

        {
            'dn': DN:
            'ChangelogDir':
            'ChangelogMaxEntries':
            'ChangelogMaxAge':
            'ChangelogCompactDBInterval':
            'ChangelogTrimInterval':
            'ChangelogMaxConcurrentWrites':
            'ChangelogEncryptionAlgorithm':
            'ChangelogSymmetricKey':
        }
    """

    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET PATCH PUT DELETE OPTIONS HEAD'
        return jsonify(msg)

    if request.method == 'HEAD':
        #
        # Return the HEAD response
        #
        try:
            ds.search_s('cn=changelog5,cn=config', ldap.SCOPE_BASE,
                        "(objectclass=*)")
            return jsonResponse("", 200, request.url)
        except:
            return jsonResponse("", 404, request.url)

    if request.method == 'GET':
        # Get changelog info
        try:
            entry = ds.search_s('cn=changelog5,cn=config',
                                ldap.SCOPE_BASE,
                                "(objectclass=*)")
            return jsonReturnChangelog(entry[0], 200, request.url)
        except ldap.NO_SUCH_OBJECT:
            msg = ("No replication changelog: " + e.message['desc'])
            return jsonResponse(msg, 404, request.url)
        except ldap.LDAPError as e:
            msg = ("Error: " + e.message['desc'])
            return jsonResponse(msg, 403, request.url)

    elif request.method == 'PATCH':
        # Modify changelog
        json_req = request.get_json(force=True)
        if not json_req:
            return jsonResponse('Request is not a valid JSON representation',
                                400, request.url)

        for mod in json_req:
            # Get the op (add, replace, delete) and parse the path
            op = jsonGetPatchOp(mod)
            if op == -1:
                return jsonResponse(('invalid PATCH op (%s)' % mod['op']), 400,
                                    reuqest.url)
            attr = str(mod['path']).lower()

            try:
                value = str(mod['value'])
                if attr == 'changelogdir':
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties={
                                                   CHANGELOG_DIR:
                                                   value})
                elif attr == 'changelogmaxage':
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties={
                                                   CHANGELOG_MAXAGE: value})
                elif attr == 'changelogmaxentries':
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties={
                                                   CHANGELOG_MAXENTRIES:
                                                   value})
                elif attr == 'changelogtriminterval':
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties={
                                                   CHANGELOG_TRIM_INTERVAL:
                                                   value})
                elif attr == 'changelogcompactdbinterval':
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties={
                                                   CHANGELOG_COMPACT_INTV:
                                                   value})
                elif attr == 'changelogmaxconcurrentwrites':
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties={
                                                   CHANGELOG_CONCURRENT_WRITES:
                                                   value})
                elif attr == 'changelogencryptionalgorithm':
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties={
                                                   CHANGELOG_ENCRYPT_ALG:
                                                   value})
                elif attr == 'changelogsymmetrickey':
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties={
                                                   CHANGELOG_SYM_KEY:
                                                   value})

                # Success, return the new change log rep, and link
                entry = ds.search_s('cn=changelog5,cn=config',
                                    ldap.SCOPE_BASE,
                                    "(objectclass=*)")
                link = '%sv1/replication/changelog' % (request.url_root)

                return jsonReturnChangelog(entry[0], 200, link)

            except ldap.LDAPError as e:
                # Failed to set changelog
                msg = ('Failed to set changelog property(%s): error %s' %
                       (str(request.args[param]), e.message['desc']))
                return jsonResponse(msg, 403)

    elif request.method == 'PUT':
        """
        Process any params::

            changelogdir                 - CHANGELOG_DIR
            changelogmaxage              - CHANGELOG_MAXAGE
            changelogmaxentries          - CHANGELOG_MAXENTRIES
            changelogtriminterval        - CHANGELOG_TRIM_INTERVAL
            changelogcompactdbinterval   - CHANGELOG_COMPACT_INTV
            changelogmaxconcurrentwrites - CHANGELOG_CONCURRENT_WRITES
            changelogencryptionalgorithm - CHANGELOG_ENCRYPT_ALG
            changelogsymmetrickey        - CHANGELOG_SYM_KEY
        """

        json_req = request.get_json(force=True)
        if not json_req:
            return jsonResponse('Request is not a valid JSON representation',
                                400, request.url)
        try:
            # Create changelog
            ds.changelog.create()
        except ldap.LDAPError as e:
            msg = ('Failed to create replication changelog: error ' + str(e))
            return jsonResponse(msg, 403)

        # Set changelog properties
        for param in json_req:
            value = str(json_req[param])
            param = param.lower()
            try:
                if param == 'changelogdir':
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties={
                                                   CHANGELOG_DIR: value})
                elif param == 'changelogmaxage':
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties={
                                                   CHANGELOG_MAXAGE: value})
                elif param == 'changelogmaxentries':
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties={
                                                   CHANGELOG_MAXENTRIES:
                                                   value})
                elif param == 'changelogtriminterval':
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties={
                                                   CHANGELOG_TRIM_INTERVAL:
                                                   value})
                elif param == 'changelogcompactdbinterval':
                    props = {CHANGELOG_COMPACT_INTV: value}
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties=props)
                elif param == 'changelogmaxconcurrentwrites':
                    props = {CHANGELOG_CONCURRENT_WRITES: value}
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties=props)
                elif param == 'changelogencryptionalgorithm':
                    props = {CHANGELOG_ENCRYPT_ALG: value}
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties=props)
                elif param == 'changelogsymmetrickey':
                    props = {CHANGELOG_SYM_KEY: value}
                    ds.changelog.setProperties(changelogdn=DN_CHANGELOG,
                                               properties=props)

                # Success, return the new change log rep, and link
                entry = ds.search_s('cn=changelog5,cn=config',
                                    ldap.SCOPE_BASE,
                                    "(objectclass=*)")
                link = '%sv1/replication/changelog' % (request.url_root)

                return jsonReturnChangelog(entry, 200, link)

            except ldap.LDAPError as e:
                # Failed to set changelog - disable change log and report error
                ds.changelog.delete()
                msg = ('Failed to set changelog property(%s): error %s' %
                       (value, e.message['desc']))
                return jsonResponse(msg, 403, request.url)

    elif request.method == 'DELETE':
        try:
            # Delete changelog
            ds.changelog.delete()
            return jsonResponse('Deleted changelog', 200, request.url)

        except ldap.LDAPError as e:
            msg = ('Failed to delete replication changelog: error ' + str(e))
            return jsonResponse(msg, 403, request.url)


@repl_api.route('/v1/replication', methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def replicationSuffixes(ds):
    """
    Return a list of backends that are replicated

    Replica Representation::

        {
            'dn': DN
            'ReplicaRoot': SUFFIX
            'ReplicaRole': master, hub, consumer
            'ReplicaType': 2
            'ReplicaFlags': 1
            'ReplicaID': 1
            'ReplicaPurgeDelay': 604800
            'ReplicaBindDN': 'uid=replica,cn=config'
            'ReplicaBindDNGroup:' DN
            'ReplicaBindDNGroupCheckInterval': 60
            'ReplicaState': <decode base64 value>
            'ReplicaName': '38475874-sdfdf-...'
            'ReplicaBackoffMin': 5
            'ReplicaBackoffMax': 300
            'ReplicaProtocolTimeout': 120
            'ReplicaReferral': REFERRAL
            'ReplicaTombstonePurgeInterval': 12000
            'ReplicaPreciseTombstonePurging': on
            'ReplicaLegacyConsumer':
            'ReplicaCleanRUV': VALUE
            'ReplicaAbortCleanRUV': VALUE
            'ReplicaChangeCount':
            'ReplicaReapActive': 0/1
        }
    """

    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    # Perform the search that HEAD and GET will use
    try:
        entries = ds.search_s('cn=config', ldap.SCOPE_SUBTREE,
                              '(objectclass=nsDS5Replica)')
    except ldap.LDAPError as e:
            return jsonResponse(str(e), 403, request.url)

    if request.method == 'HEAD':
        #
        # Return the HEAD response
        #
        if not entries:
            return jsonResponse("", 404, request.url)
        else:
            return jsonResponse("", 200, request.url)

    if request.method == 'GET':
        if entries:
            return jsonReturnReplicas(entries, 200, request.url)
        if not entries:
            return jsonResponse("", 404, request.url)

        #
        # To do MARK
        #
        # Add this to jsonReturnReplicas()
        #
        return jsonResponse("Replication State", 200, request.url)


@repl_api.route('/v1/replication/<suffix>',
                methods=['GET', 'PUT', 'PATCH', 'POST', 'DELETE', 'OPTIONS',
                         'HEAD'])
@authenticateConnection
def replicationSuffix(ds, suffix):
    """
    Replication Suffix

    Replica Representations::

        GET, PUT, PATCH
        ---------------
        {
            'dn': DN
            'ReplicaRoot': SUFFIX
            'ReplicaRole': master, hub, consumer
            'ReplicaType': 2
            'ReplicaFlags': 1
            'ReplicaID': 1
            'ReplicaPurgeDelay': 604800
            'ReplicaBindDN': 'uid=replica,cn=config'
            'ReplicaBindDNGroup:' DN
            'ReplicaBindDNGroupCheckInterval': 60
            'ReplicaState': <decode base64 value>
            'ReplicaName': '38475874-sdfdf-...'
            'ReplicaBackoffMin': 5
            'ReplicaBackoffMax': 300
            'ReplicaProtocolTimeout': 120
            'ReplicaReferral': [REFERRAL,...]
            'ReplicaTombstonePurgeInterval': 12000
            'ReplicaPreciseTombstonePurging': on
            'ReplicaLegacyConsumer':
            'ReplicaCleanRUV': VALUE
            'ReplicaAbortCleanRUV': VALUE
            'ReplicaChangeCount':
            'ReplicaReapActive': 0/1
        }

        POST - promote/demote
        ---------------------
        Replica promotion/demotion Representation
        {
            'action': 'promote/demote'
            'newrole': 'master, hub, consumer'
            'rid': 'newrid'  # Only for promotion to master
            'binddn': DN  # Only for promotion to master
            'newrole': 'master, hub, consumer'
        }
    """
    if request.method == 'OPTIONS':
        msg = {}
        msg['allow'] = 'GET PUT PATCH POST DELETE OPTIONS HEAD'
        return jsonify(msg)

    #
    # Get the replica entry, and set the response link
    #
    filter_str = ('(&(objectclass=nsDS5Replica)(nsDS5ReplicaRoot=%s))' %
                  suffix)
    try:
        replica = ds.search_s('cn=config', ldap.SCOPE_SUBTREE, filter_str)
        resp_link = '%sv1/replication/%s' % (request.url_root, suffix)
    except ldap.LDAPError as e:
        return jsonResponse("Failed to check suffix (%s)", str(e), 403,
                            request.url)

    if request.method == 'HEAD':
        if replica:
            return jsonResponse("", 200, request.url)
        else:
            return jsonResponse("", 404, request.url)

    if request.method == 'GET':
        # GET the suffix replica entry
        if not replica:
            return jsonResponse('The suffix is not enabled for replication',
                                404, request.url)
        else:
            # Return all the replication entries
            return jsonReturnReplicas(replica, 200, request.url)

    elif request.method == 'PUT':
        """
        Enable replication for this suffix
          sdf

          dfdff
        """
        replica_role = None
        replica_id = None
        props = {}

        #
        # Check if this suffix is already enabled for replication
        #
        if replica:
            # replica already exists, idempotent says to return success
            return jsonReturnReplicas(replica, 201, request.url)

        json_req = request.get_json(force=True)
        if not json_req:
            return jsonResponse('Request is not a valid JSON representation',
                                400, request.url)

        #
        # Parse the json request, and set our properties
        #
        for param in json_req:
            if param == 'dn':
                continue
            if param == 'ReplicaRoot':
                if json_req[param].lower() != suffix.lower():
                    # mismatch betwen url and rep
                    return jsonResponse(
                        'Resource (%s) does not match representation (%s)' %
                        (suffix, json_req[param].lower()),
                        403, request.url)

            elif param == 'ReplicaRole':
                replica_role = str(json_req[param])
            elif param == 'ReplicaID':
                replica_id = str(json_req[param])
            else:
                props[param] = str(json_req[param])

        #
        # Enable replication
        #
        try:
            ds.replica.enableReplication(suffix=suffix,
                                         role=replica_role,
                                         replicaId=int(replica_id),
                                         properties=props)
            try:
                replica = ds.search_s('cn=config', ldap.SCOPE_SUBTREE,
                                      filter_str)
                return jsonReturnReplicas(replica, 201, resp_link)
            except ldap.LDAPError as e:
                msg = ('Failed to get replica after creation for %s: %s' %
                       (suffix, str(e)))
                return jsonResponse(msg, 403, request.url)

        except Exception as e:
            return jsonResponse('Failed to enable replication for %s: %s' %
                                (suffix, str(e)), 403, request.url)

    elif request.method == 'POST':
        """
        Promote/Demote Replica Role::

            {
                'action': 'promote/demote'
                'newrole': 'master, hub, consumer'
                'rid': 'newrid'  # Only for promotion to master
                'binddn': DN  # Only for promotion to master
            }
        """
        #
        # Parse and validate JSON request
        #
        json_req = request.get_json(force=True)
        if not json_req:
            return jsonResponse('Request is not a valid JSON representation',
                                403, request.url)

        # Check for action and newrole
        if 'action' not in json_req or 'newrole' not in json_req:
            return jsonResponse('JSON representation incomplete',
                                403, request.url)
        else:
            action = str(json_req['action'])
            newrole = str(json_req['newrole'])

        # Check action
        if action != 'promote' and action != 'demote':
            return jsonResponse('JSON representation unknown action: ' +
                                action, 403, request.url)

        # Check role
        if newrole != 'master' and newrole != 'hub' and newrole != 'consumer':
            return jsonResponse('JSON request unknown role: ' +
                                newrole, 403, request.url)

        # Check role against replica bind dn
        if (newrole == 'hub' or newrole == 'consumer'):
            if 'binddn' not in json_req:
                return jsonResponse('JSON request missing "binddn"', 403,
                                    request.url)

        # Make sure binddn is a DN
        if 'binddn' in json_req:
            binddn = str(json_req['binddn'])
            if not is_a_dn(binddn):
                return jsonResponse('JSON request "binddn" (%s) is not a DN',
                                    binddn, 403, request.url)
        else:
            binddn = None

        #
        # Check the rid
        #
        if 'rid' in json_req:
            rid = str(json_req['rid'])
        else:
            rid = None

        #
        # Check role verse action
        #
        if action == 'promote' and newrole == 'consumer':
            return jsonResponse('Can not promote a replica to a consumer', 403,
                                request.url)
        if action == 'demote' and newrole == 'master':
            return jsonResponse('Can not demote a replica to a master', 403,
                                request.url)

        #
        # All the validation checks have been performed, now action!
        #
        if action == 'promote':
            """
            Promote a replica to a master or hub
            """
            try:
                ds.replica.promote(suffix, newrole, rid, binddn)
            except ValueError as e:
                return jsonResponse(str(e), 403, request.url)

        else:
            """
            Demote a replica to a hub or consumer
            """
            try:
                ds.replica.demote(suffix, newrole)
            except ValueError as e:
                return jsonResponse(str(e), 403, request.url)

        #
        # Success, return our json response
        #
        try:
            replica = ds.search_s('cn=config', ldap.SCOPE_SUBTREE,
                                  filter_str)
            return jsonReturnReplicas(replica, 200, resp_link)
        except ldap.LDAPError as e:
            return jsonResponse('Failed to get updated replica entry: ' +
                                str(e), 403, request.url)

    elif request.method == 'PATCH':
        # Modify replica
        mod_list = []
        json_req = request.get_json(force=True)
        if not json_req:
            return jsonResponse('Request is not a valid JSON representation',
                                403, request.url)

        for mod in json_req:
            # Get the op (add, replace, delete) and parse the path
            op = jsonGetPatchOp(mod)
            if op == -1:
                return jsonResponse(('invalid PATCH op (%s)' % mod['op']), 400,
                                    request.url)

            try:
                mod_attr = REPLICA_PROPNAME_TO_ATTRNAME[mod['path']]
            except:
                return jsonResponse('Unknown attribute (%s)' % mod['path'],
                                    403, request.url)

            if isinstance(mod['value'], list):
                value = list(map(str, mod['value']))
            else:
                value = str(mod['value'])
            mod_list.append((op, mod_attr, value))

        #
        # Perform the modifications
        #
        try:
            msg_id = ds.modify_ext(replica[0].dn, mod_list, serverctrls=[pr])
        except ldap.LDAPError as e:
            return jsonResponse('Failed to modify resource: ' + str(e), 403,
                                request.url)
        #
        # Generate the response and resource link
        #
        try:
            _, _, _, resp_ctrls = ds.result3(msg_id)
            replica = Entry(resp_ctrls[0].dn)
            replica.data = resp_ctrls[0].entry
            return jsonReturnReplicas([replica], 200, resp_link)

        except ldap.LDAPError as e:
            return jsonResponse('Failed to modify resource: ' + str(e), 403,
                                request.url)

    elif request.method == 'DELETE':
        # Delete replication config/agmts
        try:
            ds.replica.disableReplication(suffix)
            return jsonResponse('', 200, request.url)
        except ldap.LDAPError as e:
            return jsonResponse('Failed to disable replication for %s: %s' %
                                (suffix, str(e)), 403, request.url)


#
# Replication agreement resources
#
@repl_api.route('/v1/replication/<suffix>/agmts',
                methods=['GET', 'POST', 'DELETE', 'OPTIONS', 'HEAD'])
@authenticateConnection
def replAgmts(ds, suffix):
    """
    Replication Agreements

    Replication Agreement Representations::

        GET, PUT
        -------------------------------------------------------
        {
            'dn': DN
            'ReplicaName':
            'ReplicaBindMethod': 'SASL/GSSAPI, SIMPLE  '
            'ReplicaBindDN': cn=dm
            'ReplicaBindCredentials': secret
            'ReplicaHost': vm-192.abc.idm.lab.eng.brq.redhat.com
            'ReplicaPort': 389
            'ReplicaRoot': dc=abc,dc=idm,dc=lab,dc=eng,dc=brq,dc=redhat,dc=com
            'ReplicaTransportInfo': LDAP
            'ReplicatedAttributeList':
            'ReplicatedAttributeListTotal':
            'ReplicaConnTimeout': 120
            'ReplicaRefresh': start
            'ReplicaSchedule':
            'ReplicaBusyWaitTime':
            'ReplicaSessionPauseTime':
            'ReplicaEnabled':
            'ReplicaStripAttrs':
            'ReplicaFlowControlWindow':
            'ReplicaFlowControlPause':
            'ReplicaRUV': []
            'ReplicaMaxCSN':
            'ReplicaBeginRefresh':
            'ReplicaLastUpdateStart': 20151109201346Z
            'ReplicaLastUpdateEnd: 20151109201346Z
            'ReplicaChangesSentSinceStartup': base64
            'ReplicaLastUpdateStatus':
            'ReplicaUpdateInProgress':
            'ReplicaLastInitStart':
            'ReplicaLastInitEnd': 20151109201336Z
            'ReplicaLastInitStatus':
            'ReplicaReapActive':
            'ReplicaStatus':
            # Window Sync Attributes
            'WinsyncDirectoryReplicaSubtree': DN
            'WindowsReplicaSubtree': DN
            'WinsyncNewWinUserSyncEnabled': 'on/off'
            'WinsyncNewWinGroupSyncEnabled':'on/off'
            'WinsyncWindowsDomain': VALUE
            'WinsyncDirsyncCookie': BASE64-VALUE
            'WinsyncInterval': INTERVAL
            'WinsyncOneWaySync':
            'WinsyncMoveAction':
            'WinsyncWindowsFilter':
            'WinsyncDirectoryFilter':
            'WinsyncSubtreePair':
        }

        POST
        -------------------------------------------------------
        {
            'action': 'enable/disable/initialize/send_updates'
            'target': 'all (default), winsync, ds'
        }

        POST Error Response - lists agreements that were not updated
        -------------------------------------------------------
        {
            'result_status': TEXT
            'agreements': [agmt_dn, agmt_dn,...]
        }
    """

    failure_list = []
    if request.method == 'OPTIONS':
        msg = {}
        msg['allow'] = 'GET POST DELETE OPTIONS HEAD'
        return jsonify(msg)

    # Get the replica agmt entry
    try:
        filter_str = ('(|(objectclass=%s)(objectclass=%s))' %
                      (RA_OBJECTCLASS_VALUE, RA_WINDOWS_OBJECTCLASS_VALUE))
        agmts = ds.search_s('cn=config', ldap.SCOPE_SUBTREE, filter_str)
    except ldap.LDAPError as e:
        return jsonRepsonse('Failed to search configuration for agreements: %s'
                            % str(e), 403, request.url)

    if request.method == 'HEAD':
        if agmts:
            return jsonResponse("", 200, request.url)
        else:
            return jsonResponse("", 404, request.url)

    if request.method == 'GET':
        # Get agmt info/status
        if agmts:
            return jsonReturnReplAgmts(agmts, 200, request.url)
        else:
            return jsonReturnReplAgmts('No agreements found', 404, request.url)

    elif request.method == 'POST':
        '''
        Initialize, send updates, enable, disable agreement
        '''
        return doReplAgmtPOST(ds, agmts)

    elif request.method == 'DELETE':
        # Delete all the agreements
        failure_list = []
        for agmt in agmts:
            try:
                ds.delete_s(agmt.dn)
            except:
                failure_list.append(agmt.dn)

        if len(failure_list) > 0:
            resp_msg = {'result_status': 'Failed to delete all the agmts',
                        'agreements': failure_list}
            jsonRepsonse(resp_msg, 403, request.url)
        else:
            jsonRepsonse("", 200, request.url)


@repl_api.route('/v1/replication/<suffix>/agmts/<agmt>',
                methods=['GET', 'POST', 'PUT', 'PATCH',
                         'DELETE', 'OPTIONS', 'HEAD'])
@authenticateConnection
def replAgmt(ds, suffix, agmt):
    """
    Replication Agreement

    <suffix>
    <agmt> is the "cn" value of the agrrement entry

    Replication Agreement Representations::

        GET, PUT
        -------------------------------------------------------
        {
            'dn': DN
            'ReplicaName':
            'ReplicaBindMethod': 'SASL/GSSAPI, SIMPLE  '
            'ReplicaBindDN': cn=dm
            'ReplicaBindCredentials': secret
            'ReplicaHost': vm-192.abc.idm.lab.eng.brq.redhat.com
            'ReplicaPort': 389
            'ReplicaRoot': dc=abc,dc=idm,dc=lab,dc=eng,dc=brq,dc=redhat,dc=com
            'ReplicaTransportInfo': LDAP
            'ReplicatedAttributeList':
            'ReplicatedAttributeListTotal':
            'ReplicaConnTimeout': 120
            'ReplicaRefresh': start
            'ReplicaSchedule':
            'ReplicaBusyWaitTime':
            'ReplicaSessionPauseTime':
            'ReplicaEnabled':
            'ReplicaStripAttrs':
            'ReplicaFlowControlWindow':
            'ReplicaFlowControlPause':
            'ReplicaRUV': []
            'ReplicaMaxCSN':
            'ReplicaBeginRefresh':
            'ReplicaLastUpdateStart': 20151109201346Z
            'ReplicaLastUpdateEnd: 20151109201346Z
            'ReplicaChangesSentSinceStartup': base64
            'ReplicaLastUpdateStatus':
            'ReplicaUpdateInProgress':
            'ReplicaLastInitStart':
            'ReplicaLastInitEnd': 20151109201336Z
            'ReplicaLastInitStatus':
            'ReplicaReapActive':
            # Window Sync Attributes
            'WinsyncDirectoryReplicaSubtree': DN
            'WindowsReplicaSubtree': DN
            'WinsyncNewWinUserSyncEnabled': 'on/off'
            'WinsyncNewWinGroupSyncEnabled':'on/off'
            'WinsyncWindowsDomain': VALUE
            'WinsyncDirsyncCookie': BASE64-VALUE
            'WinsyncInterval': INTERVAL
            'WinsyncOneWaySync':
            'WinsyncMoveAction':
            'WinsyncWindowsFilter':
            'WinsyncDirectoryFilter':
            'WinsyncSubtreePair':
        }

        POST
        -------------------------------------------------------
        {
            'action': 'enable/disable/initialize/send_updates'
            'target': 'all (default), winsync, ds'
        }

        POST Error Response - lists agreement that were not updated
        -------------------------------------------------------
        {
            'result_status': TEXT
            'agreements': [agmt_dn, agmt_dn,...]
        }

    """

    if request.method == 'OPTIONS':
        msg = {}
        msg['allow'] = 'GET POST PUT PATCH DELETE OPTIONS'
        return jsonify(msg)

    # Get the replica agmt entry
    try:
        filter_str = ('(&(|(objectclass=%s)(objectclass=%s))(cn=%s))' %
                      (RA_OBJECTCLASS_VALUE, RA_WINDOWS_OBJECTCLASS_VALUE,
                       agmt))
        agmts = ds.search_s('cn=config', ldap.SCOPE_SUBTREE, filter_str)
        if agmts:
            agmt_entry = agmts[0]
        else:
            agmt_entry = None
    except ldap.LDAPError as e:
        return jsonRepsonse('Failed to search configuration for agreement: %s'
                            % str(e), 403, request.url)

    if request.method == 'HEAD':
        if agmt_entry:
            return jsonResponse("", 200, request.url)
        else:
            return jsonResponse("", 404, request.url)

    if request.method == 'GET':
        # Get agmt info
        if agmt_entry:
            return jsonReturnReplAgmts([agmt_entry], 200, request.url)
        else:
            return jsonResponse('Agreement not found', 404, request.url)

    elif request.method == 'POST':
        '''
        Initialize, send updates, enable, disable agreement
        '''
        return doReplAgmtPOST(ds, [agmt_entry])

    elif request.method == 'PUT':
        '''
        Create replication agmt
        '''

        #
        # Parse JSON and validate request
        #
        json_req = request.get_json(force=True)
        if not json_req:
            return jsonResponse('Request is not a valid JSON representation',
                                403, request.url)
        if 'ReplicaRoot' not in json_req:
            return jsonResponse('Request missing "ReplicaRoot"', 403,
                                request.url)
        if 'ReplicaPort' not in json_req:
            return jsonResponse('Request missing "ReplicaPort"', 403,
                                request.url)

        if 'ReplicaHost' not in json_req:
            return jsonResponse('Request missing "ReplicaHost"', 403,
                                request.url)

        #
        # Process each parameter
        #
        props = {}
        for param in json_req:
            if param == 'dn':
                continue
            if param == 'ReplicaRoot':
                if json_req[param].lower() != suffix.lower():
                    # mismatch betwen url and rep
                    return jsonResponse(
                        'Resource (%s) does not match representation (%s)' %
                        (suffix, json_req[param].lower()),
                        403, request.url)
                else:
                    props[param] = json_req[param]
            elif param == 'ReplicaName':
                if json_req[param].lower() != agmt.lower():
                    # mismatch betwen url and rep
                    return jsonResponse(
                        'Resource (%s) does not match representation (%s)' %
                        (agmt, json_req[param].lower()),
                        403, request.url)
                else:
                    props[param] = str(json_req[param])
            else:
                # Just add this parameter
                props[param] = str(json_req[param])

        if 'ReplicaName' not in json_req:
            props['ReplicaName'] = str(agmt)

        # Check if a required winsync attribute is present
        if 'WinsyncDirectoryReplicaSubtree' in json_req:
            is_winsync_agmt = True
        else:
            is_winsync_agmt = False

        #
        # Create agreement
        #
        try:
            ds.agreement.create(suffix=suffix,
                                host=str(json_req['ReplicaHost']),
                                port=str(json_req['ReplicaPort']),
                                properties=props,
                                winsync=is_winsync_agmt)
        except Exception as e:
            return jsonResponse('Failed to create repl agreement (%s)' %
                                str(e), 403, request.url)

        #
        # Get a fresh copy of the agreement and return it
        #
        try:
            filter_str = ('(&(|(objectclass=%s)(objectclass=%s))(cn=%s))' %
                          (RA_OBJECTCLASS_VALUE, RA_WINDOWS_OBJECTCLASS_VALUE,
                           agmt))
            agmts = ds.search_s('cn=config', ldap.SCOPE_SUBTREE, filter_str)
            if agmts:
                link = '%sv1/replication/%s/%s' % (request.url_root,
                                                   suffix,
                                                   agmt)
                return jsonReturnReplAgmts(agmts, 201, link)
            else:
                return jsonResponse(
                    'Failed to find agreement in configuration', 403,
                    request.url)
        except ldap.LDAPError as e:
            return jsonRepsonse(
                'Failed to search configuration for agreement: %s'
                % str(e), 403, request.url)

    elif request.method == 'PATCH':
        # Create replication agmt

        #
        # Modify/patch the index representation - CONVERT to agreement
        #
        mod_list = []
        json_rep = request.get_json(force=True)
        for mod in json_rep:
            # Get the op (add, replace, delete) and parse the path
            op = jsonGetPatchOp(mod)
            if op == -1:
                return jsonResponse('invalid PATCH op (%s)' % mod['op'], 400,
                                    request.url)

            try:
                mod_attr = RA_PROPNAME_TO_ATTRNAME[mod['path']]
            except:
                # Unknown property
                return jsonResponse('Unknown agreement attribute (%s)' %
                                    mod['path'], 400, request.url)

            if isinstance(mod['value'], list):
                value = list(map(str, mod['value']))
            else:
                value = str(mod['value'])
            mod_list.append((op, mod_attr, value))

        #
        # Perform the modifications
        #
        try:
            msg_id = ds.modify_ext(agmt_entry.dn, mod_list, serverctrls=[pr])
        except ldap.LDAPError as e:
            return jsonResponse('Failed to modify resource: ' + str(e), 403,
                                request.url)

        #
        # Generate the response and resource link
        #
        try:
            _, _, _, resp_ctrls = ds.result3(msg_id)
            resp_link = ('%s/v1/replication/%s/%s' %
                         (request.url_root, suffix, agmt))
            replagmt = Entry(resp_ctrls[0].dn)
            replagmt.data = resp_ctrls[0].entry
            return jsonReturnReplAgmts([replagmt], 200, resp_link)

        except ldap.LDAPError as e:
            return jsonResponse('Failed to modify resource: ' + str(e), 403,
                                request.url)

    elif request.method == 'DELETE':
        '''
        Delete the replication agreement
        '''
        try:
            ds.delete_s(agmt_entry.dn)
            return jsonResponse('', 200, request.url)
        except ldap.LDAPError as e:
            return jsonResponse('Failed to delete agreement: ' + str(e), 403,
                                request.url)


@repl_api.route('/v1/replication/<suffix>/agmts/status',
                methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def replStatus(ds, suffix):
    '''
    Get the status of all the replication agreements::

        {
            [
                {
                    'agreement': name,
                    'suffix': replicated suffix,
                    'consumer_host': host,
                    'consumer_port': port,
                    'status_code': 0 = in sync, -1 not in sync,
                    'status_text': 'status message'
                },
                {
                    'agreement': name,
                    ...
                }
            ]
        }
    '''
    if request.method == 'OPTIONS':
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    # Get all the replication agreements
    try:
        agmts = ds.agreement.list(suffix=suffix)
    except:
        # No agreements
        agmts = None

    if agmts is None:
        return jsonResponse("There are no replication agreements under " +
                            "suffix (%s)" % suffix, 404, request.url)

    if request.method == 'HEAD':
        if agmt_entry:
            return jsonResponse("", 200, request.url)
        else:
            return jsonResponse("", 404, request.url)

    if request.method == 'GET':
        agmt_list = []
        for agmt in agmts:
            try:
                status = ds.getReplAgmtStatus(agmt)
            except ValueError as e:
                status = str(e)
                status_code = -1
            else:
                status_code = 0
            result = {'agreement': agmt.getValue('cn'),
                      'suffix': agmt.getValue(REPL_ROOT),
                      'consumer_host': agmt.getValue(AGMT_HOST),
                      'consumer_port': int(agmt.getValue(AGMT_PORT)),
                      'status_code': status_code,
                      'status_text': status}
            agmt_list.append(result)

        return jsonResponse(agmt_list, 200, request.url)


@repl_api.route('/v1/replication/<suffix>/agmts/<agmt>/status',
                methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def replAgmtStatus(ds, suffix, agmt):
    '''
    Get the status of this replication agreement::

        {
            {
                'agreement': name,
                'suffix': replicated suffix,
                'consumer_host': host,
                'consumer_port': port,
                'status_code': 0 = in sync, -1 not in sync,
                'status_text': 'status message'
            }
        }
    '''
    if request.method == 'OPTIONS':
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    # Get the replication agreement
    try:
        agmts = ds.agreement.list(suffix=suffix)
    except:
        # No agreements
        agmts = None

    if agmts is None:
        return jsonResponse("There are no replication agreements under " +
                            "suffix (%s)" % suffix, 404)

    if request.method == 'HEAD':
        if agmt_entry:
            return jsonResponse("", 200, request.url)
        else:
            return jsonResponse("", 404, request.url)

    if request.method == 'GET':
        agmt_list = []
        for ragmt in agmts:
            if ragmt.getValue('cn') == agmt:
                # Found our agreement
                try:
                    status = ds.getReplAgmtStatus(ragmt)
                except ValueError as e:
                    status = str(e)
                    status_code = -1
                else:
                    status_code = 0
                result = {'agreement': ragmt.getValue('cn'),
                          'suffix': ragmt.getValue(REPL_ROOT),
                          'consumer_host': ragmt.getValue(AGMT_HOST),
                          'consumer_port': int(ragmt.getValue(AGMT_PORT)),
                          'status_code': status_code,
                          'status_text': status}
                agmt_list.append(result)
                break

        if len(agmt_list) == 0:
            return jsonResponse("Agreement not found", 404, request.url)
        else:
            return jsonResponse(agmt_list, 200, request.url)
