# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap
from lib389._constants import *
from lib389.properties import *
from lib389.tasks import *
from lib389.utils import *
from flask import request
from flask import Blueprint

from .Authentication import authenticateConnection
from .Helpers import jsonResponse

repl_api = Blueprint('repl_ui_api', __name__)

"""
Replication UI Resource

This resource handles the configuration attributes for the Replication Suffix,
Changelog, and Replication Agreements.
"""


@repl_ui_api.route('/v1/ui/repl/<suffix>',
                   methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def getReplValues(ds, suffix):
    '''
    Get the replication values for a single suffix.  Gather as much replication
    data as possible so we don't have to keep opening new connections for
    things like the changelog or replication agreements.

    We are expecting a JSON
    request with a list of attributes to retrieve.

    {
        'replica': [attr, attr, ...],
        'changelog': [attr, attr, ...],
        'agmt': [attr, attr, ...]
    }

    JSON Response
    {
        'replica': {'attr': [value, value], 'attr': [value], ...},
        'changelog': {'attr': [value, value], 'attr': [value], ...},
        'agmts': [
                      {'agmt': AGMT_NAME
                       'attrs':  {'attr': [value, value],
                                  'attr': [value],
                                   ...}
                      },
                      {'agmt': AGMT_NAME
                       'attrs':  {'attr': [value, value],
                                  'attr': [value],
                                   ...}
                      },
                 ]
    }
    '''

    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {}
        msg['allow'] = 'GET OPTIONS HEAD'
        return jsonify(msg)

    #
    # Perform all the searches now so we can have a proper HEAD response
    #
    replica_dn = ''
    json_resp = {}
    json_req = request.get_json(force=True)
    if not json_req:
        return jsonResponse('Request is not a valid JSON representation',
                            403, request.url)

    replica_attrs = json_req['replica']
    changelog_attrs = json_req['changelog']
    agmt_attrs = json_req['agmt']

    #
    # Get the replica attr values.  Find the replica config based on the suffix
    #
    filter_str = ("(&(objectclass=nsDS5Replica)(nsDS5Replica=" + suffix + "))")
    try:
        repl_entry = ds.search_s("cn=config", ldap.SCOPE_SUBTREE, filter_str,
                                 replica_attrs)
        if repl_entry:
            replica_dn = repl_entry[0].dn
            json_resp['replica'] = dict(repl_entry[0].data)
        else:
            json_resp['replica'] = None
    except ldap.LDAPError as e:
        msg = ('GetReplValues - Error: ' + str(e))
        return jsonResponse(msg, 403, request.url)

    #
    # Get the changelog attr values
    #
    try:
        cl_entry = ds.search_s("cn=changelog5,cn=config", ldap.SCOPE_BASE,
                               "objectclass=*", changelog_attrs)
        if cl_entry:
            json_resp['changelog'] = dict(cl_entry[0].data)
        else:
            json_resp['changelog'] = None
    except ldap.NO_SUCH_OBJECT as e:
        msg = ('GetReplValues - Error: ' + str(e))
        return jsonResponse(msg, 404, request.url)
    except ldap.LDAPError as e:
        msg = ('GetReplValues - Error: ' + str(e))
        return jsonResponse(msg, 403, request.url)

    #
    # Get all the agmt values
    #

    filter_str = "(objectclass=nsDS5ReplicationAgreement"
    agmt_attrs.append('cn')
    try:
        agmt_entries = ds.search_s(replica_dn, ldap.SCOPE_SUBTREE, filter_str,
                                   agmt_attrs)
        if agmt_entries is None:
            json_resp['agmts'] = None
    except ldap.LDAPError as e:
        msg = ('GetReplValues - Error: ' + str(e))
        return jsonResponse(msg, 403, request.url)

    # Process each agmt
    for agmt in agmt_entries:
        json_resp['agmts'].append({'agmt': agmt.getValue('cn'),
                                   'attrs': dict(repl_entry.data)})

    # Sort agmts by agmt name(if not None)
    if json_resp['agmts']:
        json_resp['agmts'].sort(key=operator.itemgetter('agmt'))

    #
    # Noe we can handle both HEAD and GET as everything is done
    #

    if request.method == 'HEAD':
        #
        # Return HEAD response
        #
        if agmt_entries is None and cl_entry is None and repl_entry is None:
            return jsonResponse("", 404, request.url)
        else:
            return jsonResponse("", 200, request.url)

    if request.methos == 'GET':
        #
        # Return our response
        #
        return jsonResponse(json_resp, 200, request.url)

